from .horizons import Vector, Observer, download
from .animation import Body, play, play2d
from .transform import mag, radec_to_altaz, sph2car, car2sph 
from .simbad import sql2df
from .time import datetime_to_jd, jd_to_datetime, utc2tdb, tdb2utc

__version__ = "2.14.0"



text = """00.00.00.00,?,Object of unknown nature
00.02.00.00,ev,transient event
01.00.00.00,Rad,Radio-source
01.02.00.00,mR,metric Radio-source
01.04.00.00,cm,centimetric Radio-source
01.06.00.00,mm,millimetric Radio-source
01.08.00.00,smm,sub-millimetric source
01.11.00.00,HI,HI (21cm) source
01.12.00.00,rB,radio Burst
01.14.00.00,Mas,Maser
02.00.00.00,IR,Infra-Red source
02.02.00.00,FIR,Far-Infrared source
02.03.00.00,MIR,Mid-Infrared source
02.04.00.00,NIR,Near-Infrared source
03.00.00.00,red,Very red source
03.03.00.00,ERO,Extremely Red Object
04.00.00.00,blu,Blue object
05.00.00.00,UV,UV-emission source
06.00.00.00,X,X-ray source
06.02.00.00,UX?,Ultra-luminous X-ray candidate
06.10.00.00,ULX,Ultra-luminous X-ray source
07.00.00.00,gam,gamma-ray source
07.03.00.00,gB,gamma-ray Burst
08.00.00.00,err,"Not an object (error, artefact, ...)"
09.00.00.00,grv,Gravitational Source
09.03.00.00,Lev,(Micro)Lensing Event
09.06.00.00,LS?,Possible gravitational lens System
09.07.00.00,Le?,Possible gravitational lens
09.08.00.00,LI?,Possible gravitationally lensed image
09.09.00.00,gLe,Gravitational Lens
09.11.00.00,gLS,Gravitational Lens System (lens+images)
09.12.00.00,GWE,Gravitational Wave Event
10.00.00.00,..?,Candidate objects
10.01.00.00,G?,Possible Galaxy
10.02.00.00,SC?,Possible Supercluster of Galaxies
10.03.00.00,C?G,Possible Cluster of Galaxies
10.04.00.00,Gr?,Possible Group of Galaxies
10.06.00.00,As?,
10.11.00.00,**?,Physical Binary Candidate
10.11.01.00,EB?,Eclipsing Binary Candidate
10.11.10.00,Sy?,Symbiotic Star Candidate
10.11.11.00,CV?,Cataclysmic Binary Candidate
10.11.11.06,No?,Nova Candidate
10.11.12.00,XB?,X-ray binary Candidate
10.11.12.02,LX?,Low-Mass X-ray binary Candidate
10.11.12.03,HX?,High-Mass X-ray binary Candidate
10.12.00.00,Pec?,Possible Peculiar Star
10.12.01.00,Y*?,Young Stellar Object Candidate
10.12.02.00,pr?,Pre-main sequence Star Candidate
10.12.02.03,TT?,T Tau star Candidate
10.12.03.00,C*?,Possible Carbon Star
10.12.04.00,S*?,Possible S Star
10.12.05.00,OH?,Possible Star with envelope of OH/IR type
10.12.07.00,WR?,Possible Wolf-Rayet Star
10.12.08.00,Be?,Possible Be Star
10.12.09.00,Ae?,Possible Herbig Ae/Be Star
10.12.11.00,HB?,Possible Horizontal Branch Star
10.12.11.02,RR?,Possible Star of RR Lyr type
10.12.11.03,Ce?,Possible Cepheid
10.12.11.07,WV?,Possible Variable Star of W Vir type
10.12.12.00,RB?,Possible Red Giant Branch star
10.12.13.00,sg?,Possible Supergiant star
10.12.13.03,s?r,Possible Red supergiant star
10.12.13.04,s?y,Possible Yellow supergiant star
10.12.13.05,s?b,Possible Blue supergiant star
10.12.14.00,AB?,Asymptotic Giant Branch Star candidate
10.12.14.01,LP?,Long Period Variable candidate
10.12.14.02,Mi?,Mira candidate
10.12.15.00,pA?,Post-AGB Star Candidate
10.12.16.00,BS?,Candidate blue Straggler Star
10.12.17.00,HS?,Hot subdwarf candidate
10.12.18.00,WD?,White Dwarf Candidate
10.12.20.00,N*?,Neutron Star Candidate
10.12.22.00,BH?,Black Hole Candidate
10.12.23.00,SN?,SuperNova Candidate
10.12.24.00,LM?,Low-mass star candidate
10.12.26.00,BD?,Brown Dwarf Candidate
12.00.00.00,mul,Composite object
12.01.00.00,reg,Region defined in the sky
12.01.05.00,vid,Underdense region of the Universe
12.02.00.00,SCG,Supercluster of Galaxies
12.03.00.00,ClG,Cluster of Galaxies
12.04.00.00,GrG,Group of Galaxies
12.04.05.00,CGG,Compact Group of Galaxies
12.05.00.00,PaG,Pair of Galaxies
12.05.05.00,IG,Interacting Galaxies
12.09.00.00,C?*,Possible (open) star cluster
12.10.00.00,Gl?,Possible Globular Cluster
12.11.00.00,Cl*,Cluster of Stars
12.11.01.00,GlC,Globular Cluster
12.11.02.00,OpC,Open (galactic) Cluster
12.12.00.00,As*,Association of Stars
12.12.01.00,St*,Stellar Stream
12.12.02.00,MGr,Moving Group
12.13.00.00,**,Double or multiple star
12.13.01.00,EB*,Eclipsing binary
12.13.01.01,Al*,Eclipsing binary of Algol type
12.13.01.02,bL*,Eclipsing binary of beta Lyr type
12.13.01.03,WU*,Eclipsing binary of W UMa type
12.13.02.00,SB*,Spectroscopic binary
12.13.05.00,El*,Ellipsoidal variable Star
12.13.10.00,Sy*,Symbiotic Star
12.13.11.00,CV*,Cataclysmic Variable Star
12.13.11.02,DQ*,CV DQ Her type (intermediate polar)
12.13.11.03,AM*,CV of AM Her type (polar)
12.13.11.05,NL*,Nova-like Star
12.13.11.06,No*,Nova
12.13.11.07,DN*,Dwarf Nova
12.13.12.00,XB*,X-ray Binary
12.13.12.02,LXB,Low Mass X-ray Binary
12.13.12.03,HXB,High Mass X-ray Binary
13.00.00.00,ISM,Interstellar matter
13.01.00.00,PoC,Part of Cloud
13.02.00.00,PN?,Possible Planetary Nebula
13.03.00.00,CGb,Cometary Globule
13.04.00.00,bub,Bubble
13.06.00.00,EmO,Emission Object
13.08.00.00,Cld,Cloud
13.08.03.00,GNe,Galactic Nebula
13.08.04.00,BNe,Bright Nebula
13.08.06.00,DNe,Dark Cloud (nebula)
13.08.07.00,RNe,Reflection Nebula
13.08.12.00,MoC,Molecular Cloud
13.08.12.03,glb,Globule (low-mass dark cloud)
13.08.12.06,cor,Dense core
13.08.12.08,SFR,Star forming region
13.08.13.00,HVC,High-velocity Cloud
13.09.00.00,HII,HII (ionized) region
13.10.00.00,PN,Planetary Nebula
13.11.00.00,sh,HI shell
13.12.00.00,SR?,SuperNova Remnant Candidate
13.13.00.00,SNR,SuperNova Remnant
13.14.00.00,cir,CircumStellar matter
13.14.01.00,of?,Outflow candidate
13.14.15.00,out,Outflow
13.14.16.00,HH,Herbig-Haro Object
14.00.00.00,*,Star
14.01.00.00,*iC,Star in Cluster
14.02.00.00,*iN,Star in Nebula
14.03.00.00,*iA,Star in Association
14.05.00.00,V*?,Star suspected of Variability
14.06.00.00,Pe*,Peculiar Star
14.06.01.00,HB*,Horizontal Branch Star
14.06.02.00,Y*O,Young Stellar Object
14.06.02.04,Ae*,Herbig Ae/Be star
14.06.05.00,Em*,Emission-line Star
14.06.05.03,Be*,Be Star
14.06.06.00,BS*,Blue Straggler Star
14.06.10.00,RG*,Red Giant Branch star
14.06.12.00,AB*,Asymptotic Giant Branch Star (He-burning)
14.06.12.03,C*,Carbon Star
14.06.12.06,S*,S Star
14.06.13.00,sg*,Evolved supergiant star
14.06.13.03,s*r,Red supergiant star
14.06.13.04,s*y,Yellow supergiant star
14.06.13.05,s*b,Blue supergiant star
14.06.14.00,HS*,Hot subdwarf
14.06.15.00,pA*,Post-AGB Star (proto-PN)
14.06.16.00,WD*,White Dwarf
14.06.17.00,LM*,Low-mass star (M<1solMass)
14.06.18.00,BD*,Brown Dwarf (M<0.08solMass)
14.06.19.00,N*,Confirmed Neutron Star
14.06.23.00,OH*,OH/IR star
14.06.25.00,pr*,Pre-main sequence Star
14.06.25.03,TT*,T Tau-type Star
14.06.30.00,WR*,Wolf-Rayet Star
14.07.00.00,PM*,High proper-motion Star
14.08.00.00,HV*,High-velocity Star
14.09.00.00,V*,Variable Star
14.09.01.00,Ir*,Variable Star of irregular type
14.09.01.01,Or*,Variable Star of Orion Type
14.09.03.00,Er*,Eruptive variable Star
14.09.03.04,RC*,Variable Star of R CrB type
14.09.03.05,RC?,Variable Star of R CrB type candiate
14.09.04.00,Ro*,Rotationally variable Star
14.09.04.01,a2*,Variable Star of alpha2 CVn type
14.09.04.03,Psr,Pulsar
14.09.04.04,BY*,Variable of BY Dra type
14.09.04.05,RS*,Variable of RS CVn type
14.09.05.00,Pu*,Pulsating variable Star
14.09.05.02,RR*,Variable Star of RR Lyr type
14.09.05.03,Ce*,Cepheid variable Star
14.09.05.05,dS*,Variable Star of delta Sct type
14.09.05.06,RV*,Variable Star of RV Tau type
14.09.05.07,WV*,Variable Star of W Vir type
14.09.05.08,bC*,Variable Star of beta Cep type
14.09.05.09,cC*,Classical Cepheid (delta Cep type)
14.09.05.10,gD*,Variable Star of gamma Dor type
14.09.05.11,SX*,Variable Star of SX Phe type (subdwarf)
14.09.06.00,LP*,Long-period variable star
14.09.06.01,Mi*,Variable Star of Mira Cet type
14.09.08.00,SN*,SuperNova
14.14.00.00,su*,Sub-stellar object
14.14.02.00,Pl?,Extra-solar Planet Candidate
14.14.10.00,Pl,Extra-solar Confirmed Planet
15.00.00.00,G,Galaxy
15.01.00.00,PoG,Part of a Galaxy
15.02.00.00,GiC,Galaxy in Cluster of Galaxies
15.02.02.00,BiC,Brightest galaxy in a Cluster (BCG)
15.03.00.00,GiG,Galaxy in Group of Galaxies
15.04.00.00,GiP,Galaxy in Pair of Galaxies
15.05.00.00,HzG,Galaxy with high redshift
15.07.00.00,rG,Radio Galaxy
15.08.00.00,H2G,HII Galaxy
15.09.00.00,LSB,Low Surface Brightness Galaxy
15.10.00.00,AG?,Possible Active Galaxy Nucleus
15.10.07.00,Q?,Possible Quasar
15.10.11.00,Bz?,Possible Blazar
15.10.17.00,BL?,Possible BL Lac
15.11.00.00,EmG,Emission-line galaxy
15.12.00.00,SBG,Starburst Galaxy
15.13.00.00,bCG,Blue compact Galaxy
15.14.00.00,LeI,Gravitationally Lensed Image
15.14.01.00,LeG,Gravitationally Lensed Image of a Galaxy
15.14.07.00,LeQ,Gravitationally Lensed Image of a Quasar
15.15.00.00,AGN,Active Galaxy Nucleus
15.15.01.00,LIN,LINER-type Active Galaxy Nucleus
15.15.02.00,SyG,Seyfert Galaxy
15.15.02.01,Sy1,Seyfert 1 Galaxy
15.15.02.02,Sy2,Seyfert 2 Galaxy
15.15.03.00,Bla,Blazar
15.15.03.01,BLL,BL Lac - type object
15.15.03.02,OVV,Optically Violently Variable object
15.15.04.00,QSO,Quasar"""



from datetime import datetime, timedelta
import math

d2r = math.pi/180
LP = 37


def sec_to_jd(seconds):
    """seconds since J2000 to jd"""
    return 2451545.0 + seconds / 86400.0


def jd_to_sec(jd):
    """jd to seconds since J2000"""
    return (jd - 2451545.0) * 86400.0

  
def datetime_to_jd(t):
    year = t.year
    month = t.month
    t_d = t.day
    t_H = t.hour
    t_M = t.minute
    t_S = t.second
    t_MS = t.microsecond
    day = t_d + t_H/24 + t_M/(24*60) + t_S/(24*60*60) + t_MS/(24*60*60*1000000)
    
    if month == 1 or month == 2:
        yearp = year - 1
        monthp = month + 12
    else:
        yearp = year
        monthp = month
    
    if ((year < 1582) or
        (year == 1582 and month < 10) or
        (year == 1582 and month == 10 and day < 15)):
        # before start of Gregorian calendar
        B = 0
    else:
        # after start of Gregorian calendar
        A = math.trunc(yearp / 100.)
        B = 2 - A + math.trunc(A / 4.)
        
    if yearp < 0:
        C = math.trunc((365.25 * yearp) - 0.75)
    else:
        C = math.trunc(365.25 * yearp)
        
    D = math.trunc(30.6001 * (monthp + 1))
    jd = B + C + D + day + 1720994.5
    
    return jd


def jd_to_datetime(jd):

    jd = jd + 0.5
    F, I = math.modf(jd)
    I = int(I)
    A = math.trunc((I - 1867216.25)/36524.25)
    
    if I > 2299160:
        B = I + 1 + A - math.trunc(A / 4.)
    else:
        B = I
        
    C = B + 1524
    D = math.trunc((C - 122.1) / 365.25)
    E = math.trunc(365.25 * D)
    G = math.trunc((C - E) / 30.6001)
    day = C - E + F - math.trunc(30.6001 * G)
    
    if G < 13.5:
        month = G - 1
    else:
        month = G - 13
        
    if month > 2.5:
        year = D - 4716
    else:
        year = D - 4715

    d_ = day
    d = int(d_)
    h_ = (d_-d)*24
    h = int(h_)
    m_ = (h_-h)*60
    m = int(m_)
    s_ = (m_-m)*60
    s = int(s_)
    ms_ = (s_-s)*1000000
    ms = int(ms_)
    dt = datetime(year, month, d, h, m, s, ms)
        
    return dt
  

def utc2tt(t, lp=LP):
    TAI = t + timedelta(seconds=lp)
    TT = TAI + timedelta(seconds=32.184)
    return TT

def tt2utc(t, lp=LP):
    TAI = t - timedelta(seconds=32.184)
    utc = TAI - timedelta(seconds=lp)
    return utc

def tt2tdb(t, lp=LP):
    T = (datetime_to_jd(t) - 2451545) / 36525 # centuries from J2000
    g = (2*math.pi/360) * (357.528 + 35999.050*T)
    dt = 0.001658 * math.sin(g + 0.0167*math.sin(g))
    TDB = t + timedelta(seconds=dt)
    return TDB

def tdb2tt(tdb):
    g = 357.53 + 0.9856003*(datetime_to_jd(tdb)-2451545)
    dt = (0.001658*math.sin(g*d2r) + 0.000014*math.sin(2*g*d2r))
    tt = tdb - timedelta(seconds=dt)
    return tt

def utc2tdb(utc, lp=LP):
    tt = utc2tt(utc, lp)
    tdb = tt2tdb(tt, lp)
    return tdb

def tdb2utc(tdb, lp=LP):
    tt = tdb2tt(tdb)
    utc = tt2utc(tt, lp=lp)
    return utc



import numpy as np
from .time import utc2tt, datetime_to_jd
from datetime import datetime

d2r = np.pi/180
r2d = 180/np.pi

def rev(x):
    if x >= 360:
        while x >= 360:
            x = x - 360
    elif x < 0:
        while x < 0:
            x = 360 + x
    return x


def get_sun(t):
    """
    Get geocentric position of the Sun

    Approximation based on method presented by The Astronomical Almanac,
    to calculate the apparent coordinates of the Sun, mean equinox and
    ecliptic of date, to a precision of about 0°.01 (36″), for dates 
    between 1950 and 2050.
    Ref: https://en.wikipedia.org/wiki/Position_of_the_Sun

    Parameters
    ----------
        t (datetime): time of observation (UTC)

    Returns
    -------
        ra  : RA of the Sun (deg)
        dec : DEC of the Sun (deg)
        r   : Distance of the Sun (AU)
    """
    tt = utc2tt(t)
    jd = datetime_to_jd(tt)
    n = jd - 2451545
    L = rev(280.460 + 0.985647*n) # mean longitude
    g = rev(357.528 + 0.9856003*n) # mean anomaly
    # Ecliptic coordinates
    ecl_lon = rev(L + 1.915 * np.sin(g*d2r) + 0.020 * np.sin(2*g*d2r))
    ecl_lat = 0
    r = 1.00014 - 0.01671*np.cos(g*d2r) - 0.00014*np.cos(2*g*d2r)
    # Equatorial coordinates
    obl = 23.439 - 0.0000004*n
    ra = rev(np.arctan2(np.cos(obl*d2r)*np.sin(ecl_lon*d2r), np.cos(ecl_lon*d2r))*r2d)
    dec = np.arcsin(np.sin(obl*d2r)*np.sin(ecl_lon*d2r))*r2d
    return ra, dec, r


def dec_sun(t):
    t = utc2tt(t)
    t0 = datetime(t.year, 1, 1)
    dt = t - t0
    N = dt.days + dt.seconds/86400 + dt.microseconds/(1_000_000*86400)
    in_cos = 0.98565*d2r*(N+10) + 1.914*d2r * np.sin(0.98565*d2r*(N-2))
    dec = -np.arcsin(0.39779 * np.cos(in_cos)) * r2d
    return dec



"""
Module transform
================
This module supplies classes and functions for dealing with coordinates
"""
import numpy as np
from datetime import datetime, timedelta
from collections.abc import Iterable
import re
from .time import datetime_to_jd
from .iau2000A import gcrs2tete

d2r = np.pi/180
r2d = 180/np.pi

def _rev(x):
    if x >= 360:
        while x >= 360:
            x = x - 360
    elif x < 0:
        while x < 0:
            x = 360 + x
    return x

rev = np.vectorize(_rev)


class RAhms:
    def __init__(self, h=0, m=0, s=0):
        self.h = h
        self.m = m
        self.s = s
        self.hours = h + m/60 + s/3600
        self.minutes = h*60 + m + s/60
        self.seconds = h*3600 + m*60 + s
        self.deg = 15 * self.hours
        self.rad = self.deg * (3.141592653589793/180)
        

class DECdms:
    def __init__(self, sign='+', d=0, m=0, s=0):
        self.sign = sign
        sgn = -1 if sign=='-' else 1
        self.d = d
        self.m = m
        self.s = s
        self.degrees = (d + m/60 + s/3600) * sgn
        self.minutes = (d*60 + m + s/60) * sgn
        self.seconds = (d*3600 + m*60 + s) * sgn
        self.deg = self.degrees
        self.rad = self.deg * (3.141592653589793/180)


class RAdeg:
    def __init__(self, ra):
        self.rad = ra * (3.141592653589793/180)
        self.h = int(ra/15)
        self.m = int(((ra/15)-self.h)*60)
        self.s = ((((ra/15)-self.h)*60)-self.m)*60
        self.str = str(self.h)+'h' + str(self.m)+'m' + str(self.s)+'s'
        

class DECdeg:
    def __init__(self, dec):
        self.rad = dec * (3.141592653589793/180)
        self.sign = '-' if dec<0 else '+'
        dec = abs(dec)
        self.d = int(dec)
        self.m = abs(int((dec-self.d)*60))
        self.s = (abs((dec-self.d)*60)-self.m)*60
        self.str = self.sign + str(self.d)+'d' + \
                   str(self.m)+'m' + str(self.s)+'s'
        

def to_epoch(ra, dec, epoch):
    """
    Convert coordinates to another epoch

    Note: you have to multiply d_ra and d_dec by (epoch-2000) and
    add the results to the initial ra and dec to get the new ra and
    dec in the desired epoch.

    Example:
    >>> ra2000, dec2000 = 11.795879583333333, -26.3824475
    >>> epoch = 1950
    >>> d_ra, d_dec = to_epoch(ra2000, dec2000, epoch=epoch)
    >>> ra = ra2000 + (epoch-2000)*d_ra
    >>> dec = dec2000 + (epoch-2000)*d_dec
    >>> print(ra,dec)
    11.183690048723463 -26.65500294330134
    
    Arguments
    ---------
        ra  (deg)    : ra in J2000
        dec (deg)    : dec in J2000
        epoch (year) : epoch to convert to
        
    Returns
    -------
        d_ra  : delta RA per year (deg/yr)
        d_dec : delta DEC per year (deg/yr)
    """
    T = (epoch - 2000) / 100
    m = 3.07496 + 0.00186*T
    n = 1.33621 - 0.00057*T
    nn = 20.0431 - 0.0085*T

    d_ra = m + n*np.sin(ra*d2r)*np.tan(dec*d2r)
    d_dec = nn*np.cos(ra*d2r)

    d_ra = RAhms(s=d_ra).deg
    if d_dec<0:
        d_dec = DECdms(sign='-', s=d_dec).deg
    else:
        d_dec = DECdms(s=d_dec).deg
    return d_ra, d_dec


def to_tete(pos, t):
    """
    Convert GCRS J2000 coordinates to True Equator True Equinox (of date)
    
    Arguments
    ---------
        pos  (np.array) : [x, y, z] coordinates in GCRS J2000
        t (datetime)    : Observation time (UTC)
        
    Returns
    -------
        position at True Equator True Equinox
    """
    return gcrs2tete(pos, t)


def obliq(t):
    obl_cf = [23.439279444444445, -0.013010213611111111,
              -5.0861111111111115e-08, 5.565e-07,
              -1.6e-10, -1.2055555555555555e-11]
    T = (datetime_to_jd(t) - 2451545) / 36525 # centuries from J2000
    cf = np.array(obl_cf[::-1])
    f = np.poly1d(cf)
    return f(T)


def _time(t):
    if isinstance(t, datetime):
        return t
    elif isinstance(t, str) and bool(re.match("\d{4}-\d\d-\d\d \d\d:\d\d:\d\d", t)):
        return datetime.strptime(t, '%Y-%m-%d %H:%M:%S')
    elif t is None:
        return datetime.utcnow()
    else:
        raise Exception("only datetime or str: '%Y-%m-%d %H:%M:%S'")
   

def mag(x):
    """Returns magnitude of a vector"""
    return np.linalg.norm(np.array(x))

def unit(x):
    """Returns unit vector of a vector"""
    return x / mag(x)

def to_xy_plane(pos):
    """Returns new positions transformed in the xy plane"""
    mags = np.array([mag(i) for i in pos])
    pos_z0 = pos * np.array([1,1,0])
    u = np.array([unit(i) for i in pos_z0])
    new_pos = np.array([u[i] * mags[i] for i in range(len(mags))])
    return new_pos

def rotating_coords(pos, period, times):
    """
    Transforms plane position coordinates (with z=0) to rotating frame
    
    Arguments
    ---------
        pos (np.array): 3d position array
        perios (float): period in days
        times (list): times list
        
    Returns
    -------
        rotating frame position array
    """
    # calculating theta
    t = np.array(times) - times[0]
    t = np.array([i.total_seconds() for i in t])
    period = period * 86400
    omega = 2*np.pi / period
    theta = omega * t

    # calculating rotating coordinates
    x, y = pos[:, 0], pos[:, 1]
    rot_x = x*np.cos(-theta) - y*np.sin(-theta)
    rot_y = x*np.sin(-theta) + y*np.cos(-theta)
    rot_pos = np.c_[rot_x, rot_y, np.zeros(len(theta))]
    return rot_pos


def radec_to_altaz(ra, dec, obs_loc, t):
    """
    Convert ra/dec coordinates to az/alt coordinates

    Arguments
    ---------
        lon (float): longtitude of observer location
        lat (float): latitude of observer location
        ra (iter of float): right ascension value(s)
        dec (iter of float): declination value(s)
        t (datetime or str): time of observation in UTC. default now.

    Returns
    -------
        altitude(s), azimuth(s)
    """
    lon, lat = obs_loc
    t = _time(t)
    if isinstance(ra, Iterable):
        ra = np.array(ra)
        dec = np.array(dec)

    J2000 = datetime(2000,1,1,12)
    d = (t - J2000).total_seconds() / 86400 #day offset

    UT = t.hour + t.minute/60 + t.second/3600
    LST = (100.46 + 0.985647 * d + lon + 15*UT + 360) % 360
    ha = (LST - ra + 360) % 360
    
    x = np.cos(ha*d2r) * np.cos(dec*d2r)
    y = np.sin(ha*d2r) * np.cos(dec*d2r)
    z = np.sin(dec*d2r)
    xhor = x*np.cos((90-lat)*d2r) - z*np.sin((90-lat)*d2r)
    yhor = y
    zhor = x*np.sin((90-lat)*d2r) + z*np.cos((90-lat)*d2r)
    az = np.arctan2(yhor, xhor)*r2d + 180
    alt = np.arcsin(zhor)*r2d
    return az, alt


def altaz_to_radec(lon, lat, az, alt, t=None):
    """
    Convert az/alt coordinates to ra/dec coordinates

    Arguments
    ---------
        lon (float): longtitude of observer location
        lat (float): latitude of observer location
        az (iter of float): azimuth value(s)
        alt (iter of float): altitude value(s)
        t (datetime or str): time of observation in UTC. default now.

    Returns
    -------
        right ascension value(s), declination value(s)
    """

    t = _time(t)
    if isinstance(az, Iterable):
        az = np.array(az)
        alt = np.array(alt)

    ALT = alt * d2r
    AZ  = az  * d2r
    LAT = lat * d2r

    sin_DEC = np.sin(LAT)*np.sin(ALT) + np.cos(LAT)*np.cos(ALT)*np.cos(AZ)
    DEC = np.arcsin(sin_DEC)

    cos_HA = (np.sin(ALT) - np.sin(LAT)*np.sin(DEC)) / \
             (np.cos(LAT)*np.cos(DEC))
    
    HA = np.where(cos_HA>1, 0, np.arccos(cos_HA))

    dec = DEC * r2d
    ha  = HA  * r2d
    
    J2000 = datetime(2000,1,1,12)
    d = (t - J2000).total_seconds() / 86400
    UT = t.hour + t.minute/60 + t.second/3600
    LST = (100.46 + 0.985647 * d + lon + 15*UT + 360) % 360

    ra = np.where(az>=180, (LST-ha)%360, (LST+ha)%360)

    return ra, dec


def _in_vec(inp):
    if len(inp.shape)>1:
        if inp.shape[-1]==3:
            vec_out = inp[:,0], inp[:,1], inp[:,2]
        elif inp.shape[-1]==2:
            vec_out = inp[:,0], inp[:,1]
        else:
            raise Exception('Input imension not sopported!')
    elif len(inp.shape)==1:
        vec_out = inp
    else:
        raise Exception('Dimensionality problem! 2')
    return vec_out


def _out_vec(vec):
    if len(vec.shape)>1:
        vec = vec.T
        if vec.shape[-1]==3:
            x, y, z = vec[:,0], vec[:,1], vec[:,2]
            vec_out = np.vstack((x, y, z)).T
        elif vec.shape[-1]==2:
            x, y = vec[:,0], vec[:,1]
            vec_out = np.vstack((x, y)).T
        else:
            raise Exception('Dimensionality problem! 3')
    elif len(vec.shape)==1:
        vec_out = vec
    else:
        raise Exception('Dimensionality problem! 4')
    return vec_out
        


def sph2car(sph):
    lon, lat, r = _in_vec(sph)
    lon, lat = lon*d2r, lat*d2r
    x = r * np.cos(lat) * np.cos(lon)
    y = r * np.cos(lat) * np.sin(lon)
    z = r * np.sin(lat)
    car = _out_vec(np.array([x, y, z]))
    return car



def car2sph(car):
    x, y, z = _in_vec(car)
    lon = rev(np.arctan2(y, x)*r2d)
    lat = np.arctan2(z, np.sqrt(x**2 + y**2))*r2d
    r = np.sqrt(x**2 + y**2 + z**2)
    sph = _out_vec(np.array([lon, lat, r]))
    return sph


def _equsph2eclsph(radec):
    ra, dec = _in_vec(radec)
    ra, dec = ra*d2r, dec*d2r
    epsilon = (23 + 26/60 + 21.448/3600)*d2r
    sb = np.sin(dec)*np.cos(epsilon) - np.cos(dec)*np.sin(epsilon)*np.sin(ra)
    cbcl = np.cos(dec)*np.cos(ra)
    cbsl = np.sin(dec)*np.sin(epsilon) + np.cos(dec)*np.cos(epsilon)*np.sin(ra)
    lon = rev(np.arctan2(cbsl,cbcl)*r2d)
    r = np.sqrt(cbsl**2 + cbcl**2)
    lat = np.arctan2(sb,r)*r2d
    lonlat = _out_vec(np.array([lon, lat]))
    return lonlat


def equ_sph2ecl_sph(equsph):
    if len(equsph.shape)>1:
        eclsph = np.zeros((len(equsph), 3))
        eclsph[:,:2] = _equsph2eclsph(equsph[:,:2])
        eclsph[:, 2] = equsph[:,2]
    else:
        lon, lat = _equsph2eclsph(equsph[:2])
        eclsph = np.array([lon, lat, equsph[2]])
    return eclsph


def equ_car2ecl_car(equ_car):
    equ_sph = car2sph(equ_car)
    ecl_sph = equ_sph2ecl_sph(equ_sph)
    ecl_car = sph2car(ecl_sph)
    return ecl_car



def angular_separation(r1, d1, r2, d2):
    """
    Calculate angular separation between two point

    Arguments
    ---------
        r1 (float): right ascension of the first point in degrees
        d1 (float): declination of the first point in degrees
        r2 (float): right ascension of the second point in degrees
        d2 (float): declination of the second point in degrees

    Returns
    -------
        angular sepration in degrees
    """
    r1 = (np.pi/180) * r1
    d1 = (np.pi/180) * d1
    r2 = (np.pi/180) * r2
    d2 = (np.pi/180) * d2

    radical = np.sqrt(
        (np.cos(d2)**2)*(np.sin(r2-r1)**2) + ((np.cos(d1)*np.sin(d2) - np.sin(d1)*np.cos(d2)*np.cos(r2-r1))**2)
        )
    
    kasr = radical / ( (np.sin(d1)*np.sin(d2)) + (np.cos(d1)*np.cos(d2)*np.cos(r2-r1)) )

    sep = (180/np.pi) * np.arctan(kasr)
    
    return sep


def posvel(ra, dec, pmra, pmdec, distance, radvel):
    """
    Cartesian position and velocity vectors.
    
    Note:
    -----
    1) If you have parallax instead of distance, use this formulla:
    distance = (1/(plx/1000)) * 30856775814913.67 # km
    
    2) If you have redshift instead of radial velocity, use this formulla:
    radvel = z * 299792.458 # km/s
    
    Arguments
    ---------
        ra (iter of float): RA of objects
        dec (iter of float): DEC of objects
        pmra (iter of float): pmRA of objects
        pmdec (iter of float): pmDEC of objects
        distance (iter of float): Distance to objects in km
        radvel (iter of float): Radial Velocity of objects in km/s
        
    Returns
    -------
        position vectors (km), velocity vectors (km/s)
    """
    
    # Changes of RA and DEC in one second
    d_ra = ((pmra/1000)/3600)/31557600
    d_dec = ((pmdec/1000)/3600)/31557600
    
    r1, d1 = ra, dec
    r2 = r1 + d_ra
    d2 = d1 + d_dec
    
    # Changes in distance
    dist1 = distance
    dist2 = dist1 + radvel
    
    # Convert to cartesian
    x1, y1, z1 = radec_to_cartesian(r1, d1, dist1)
    x2, y2, z2 = radec_to_cartesian(r2, d2, dist2)
    pos1 = np.array([x1,y1,z1])
    pos2 = np.array([x2,y2,z2])
    
    # Velocity
    d_pos = pos2 - pos1
    vel = d_pos * 1 # because it's in one second
    return pos1, vel



"""
Module simbad
=============
This module deals with deep-sky objects by querying SIMBAD astronomical
database from Centre de données astronomiques de Strasbourg (CDS).

Note about otype
----------------
Several functions in this module have a parameter 'otype'. This is the
type of deep-sky object defined by SIMBAD in this url:
http://simbad.u-strasbg.fr/simbad/sim-display?data=otypes

This a hierarchical classification. For exact type of an object, you
shoud enter its numeric code as a string, such as '14.09.05.03' for
Cepheid variable Star. You can also enter '14.09.05' for all Pulsating
variable Stars (including cepheids and other Pulsating variable Stars
subclasses). If you enter '14.09.00.00' or '14.09' (they are same), it
means all types of variable stars.

For simplicity, we have also provided 7 string values ('radiation',
'gravitation', 'candidate', 'multiple', 'interstellar', 'star' and
'galaxy'). Theses are general categories which encompass many subcategories. 
"""
import numpy as np
import pandas as pd
from urllib.request import urlopen
import requests
import json, csv, re
from io import StringIO
from datetime import datetime
from .simbad_otypes import text
from .transform import altaz_to_radec, radec_to_altaz


data = StringIO(text)

code, short, long = [], [], []

for row in csv.reader(data):
    code.append(row[0])
    short.append(row[1])
    long.append(row[2])

_dc = {k:v for (k,v) in zip(code, short)}
otype_dict = {k:v for (k,v) in zip(code, long)}

def object_type(num):
    
    if num[0].isalpha():
        if num.lower()=='radiation':
            num = '999'
        elif num.lower()=='gravitation':
            num = '09'
        elif num.lower()=='candidate':
            num = '10'
        elif num.lower()=='multiple':
            num = '12'
        elif num.lower()=='interstellar':
            num = '13'
        elif num.lower()=='star':
            num = '14'
        elif num.lower()=='galaxy':
            num = '15'
            
    if bool(re.match("\d\d.\d\d.\d\d.\d\dZ", num+'Z')):
        otypes = [_dc[num]]
    elif bool(re.match("\d\d.\d\d.\d\dZ", num+'Z')):
        otypes = [_dc[i] for i in _dc.keys() if i[:8]==num]
    elif bool(re.match("\d\d.\d\dZ", num+'Z')):
        otypes = [_dc[i] for i in _dc.keys() if i[:5]==num]
    elif bool(re.match("\d\dZ", num+'Z')):
        otypes = [_dc[i] for i in _dc.keys() if i[:2]==num]
    elif num=='999':
        otypes = ['Rad','mR','cm','mm','smm','HI','rB','Mas','IR','FIR',
                  'MIR','NIR','red','ERO','blu','UV','X','UX?','ULX',
                  'gam', 'gB']
    else:
        raise Exception('Object type invalid!')
            
    return otypes


def bright_objects(otype=None):
    """
    Return Deep-sky objects with apparent magnitudes less than 4.5

    Arguments
    ---------
        otype (str): type of object. Default None of all objects

    Returns
    -------
        Two lists : data field names, rows of data
    """
    
    from .simbad_bright_objects import mag45
    f_mag45 = StringIO(mag45)
    rows_mag45 = []
    for row in csv.reader(f_mag45):
        rows_mag45.append(row)
    rows = [[i[0], i[1], float(i[2]), float(i[3]), float(i[4])] \
          for i in rows_mag45]
    if otype is not None:
        otype = object_type(otype)
        rows = [i for i in rows if i[1] in otype]
    field_names = ['main_id', 'otype_txt', 'V', 'ra', 'dec']
    return field_names, rows
    


BASE_SIMBAD = 'http://simbad.u-strasbg.fr/simbad/sim-tap/sync?\
request=doQuery&lang=adql&format=json&query='


def get_objects(otype=None, n_max=1000):
    """
    Return deep-sky objects

    Arguments
    ---------
        otype (str): type of object. Default None of all objects
        n_max (int): maximum number of rows to return; default 1000

    Returns
    -------
        Two lists : data field names, rows of data
    """
    
    if otype is not None:
        otype = str(tuple(object_type(otype)))
        add_otype = f" WHERE otype_txt in {otype}"
    else:
        add_otype = ""
    
    sql = f"""SELECT TOP {n_max} main_id, otypedef.otype_longname, allfluxes.V, ra, dec
    FROM basic LEFT JOIN allfluxes ON basic.oid=allfluxes.oidref
    LEFT JOIN otypedef ON basic.otype=otypedef.otype"""+add_otype+" ORDER BY V"
    sql = ' '.join([i.strip() for i in sql.split('\n')])
    url = (BASE_SIMBAD+sql).replace(' ', '%20')
    r = json.loads(urlopen(url).read().decode('utf-8'))
    field_names = [dc['name'] for dc in r['metadata']]
    rows = r['data']
    return field_names, rows


def explore_region(ra, dec, r, n_max=1000):
    """
    Explore a circular region in the sky with ICRS coordinates

    Arguments
    ---------
        ra (float): right ascension of the cirular region
        dec (float): declination of the cirular region
        r (float): radius of the cirular region in degrees
        n_max (int): maximum number of rows to return; default 1000

    Returns
    -------
        Two lists : data field names, rows of data
    """
    return search_region(ra=ra, dec=dec, r=r, otype=None, n_max=n_max)

def search_region(ra, dec, r, otype=None, n_max=1000):
    """
    Explore a circular region in the sky with ICRS coordinates

    Arguments
    ---------
        ra (float): right ascension of the cirular region
        dec (float): declination of the cirular region
        r (float): radius of the cirular region in degrees
        otype (str): type of object. Default None of all objects
        n_max (int): maximum number of rows to return; default 1000

    Returns
    -------
        Two lists : data field names, rows of data
    """
    if otype is not None:
        otype = str(tuple(object_type(otype)))
        add_otype = f" AND otype_txt in {otype}"
    else:
        add_otype = ""
    sql = f"""SELECT TOP {n_max} main_id, otypedef.otype_longname, allfluxes.V, ra, dec
    FROM basic LEFT JOIN allfluxes ON basic.oid=allfluxes.oidref
    LEFT JOIN otypedef ON basic.otype=otypedef.otype
    WHERE CONTAINS(POINT('ICRS', ra, dec), CIRCLE('ICRS', {ra}, {dec}, {r})) = 1
    AND ra IS NOT NULL AND dec IS NOT NULL"""+add_otype+" ORDER BY V"
    sql = ' '.join([i.strip() for i in sql.split('\n')])
    url = (BASE_SIMBAD+sql).replace(' ', '%20')
    r = json.loads(urlopen(url).read().decode('utf-8'))
    field_names = [dc['name'] for dc in r['metadata']]
    rows = r['data']
    return field_names, rows

def explore_sky(lon, lat, t=None, az=0, alt=90, r=1, n_max=1000):
    """
    Explore a circular region in the local sky with AltAz coordinates

    Arguments
    ---------
        lon (float): longtitude of observer location
        lat (float): latitude of observer location
        t (datetime or str): time of observation in UTC. default now.
        az (float): azimuth of center of the circle
        alt (float): altitude of center of the circle
        r (float): radius of the cirular region in degrees
        n_max (int): maximum number of rows to return; default 1000

    Returns
    -------
        Two lists : data field names, rows of data
    """
    return search_sky(lon, lat, t, az, alt, r, otype=None, n_max=1000)

def search_sky(lon, lat, t=None, az=0, alt=90, r=1, otype=None, n_max=1000):
    """
    search in a circular region in the local sky with AltAz coordinates

    Arguments
    ---------
        lon (float): longtitude of observer location
        lat (float): latitude of observer location
        t (datetime or str): time of observation in UTC. default now.
        az (float): azimouth of center of the circle
        alt (float): altitude of center of the circle
        r (float): radius of the cirular region in degrees
        otype (str): type of object. Default None of all objects
        n_max (int): maximum number of rows to return; default 1000

    Returns
    -------
        Two lists : data field names, rows of data
    """

    if t is None:
        t = datetime.utcnow()
    elif isinstance(t, datetime):
        pass
    elif isinstance(t, str) and bool(re.match("\d{4}-\d\d-\d\d \d\d:\d\d:\d\d", t)):
        t = datetime.strptime(t, '%Y-%m-%d %H:%M:%S')
    else:
        raise Exception("t should be a datetime or str: '%Y-%m-%d %H:%M:%S'")
    
    ra, dec = altaz_to_radec(lon, lat, az, alt, t)
    return search_region(ra=ra, dec=dec, r=r, otype=otype, n_max=n_max)

def sql2df(script):
    BASE = 'http://simbad.u-strasbg.fr/simbad/sim-tap/sync?request=doQuery&lang=adql&format=csv&query='
    script = ' '.join(script.strip().split('\n'))
    url = BASE+script.replace(' ', '%20') + '&format=csv'
    req = requests.request('GET', url)
    r = req.content.decode('utf-8')
    lines = r.splitlines()
    col = lines[0].split(',')
    data_lines = [i.split(',') for i in lines[1:]]
    return pd.DataFrame(data_lines, columns=col)
